unit l_gl;
(*******************************)
(* TRacer graphic libary       *)
(*******************************)
(* Created by XProger          *)
(* begin : 16.04.2005          *)
(* end   : 19.04.2005          *)
(*******************************)
(* mail : XProger@list.ru      *)
(*******************************)
interface

uses
 Windows, OpenGL,
 l_math, l_utils;

procedure glGenTextures(n: GLsizei; textures: PGLuint); stdcall; external opengl32;
procedure glBindTexture(target: GLenum; texture: GLuint); stdcall; external opengl32;
procedure glDeleteTextures(N: GLsizei; Textures: PGLuint); stdcall; external opengl32;

const
 GL_CLAMP_TO_EDGE = $812F;
 GL_RGB8          = $8051;
 GL_RGBA8         = $8058;
 GL_TEXTURE0      = $84C0;
 GL_TEXTURE1      = $84C1;

type
 TTexture = DWORD;

var
 wglSwapIntervalEXT    : function(interval: GLint): Boolean; stdcall;
 wglGetSwapIntervalEXT : function: GLint; stdcall;
 glActiveTexture       : procedure(texture: GLuint); stdcall;
 glMultiTexCoord2fv    : procedure(target: GLuint; const v: PGLfloat); stdcall;

 Initialized : boolean = false;

 Handle : HWND;
 DC     : HDC;
 RC     : HGLRC;
 isQuit : boolean = false;

 glWidth      : integer = {1280;//}640;
 glHeight     : integer = {1024;//}480;
 glBPP        : integer = 16;
 glFreq       : integer = 85;
 glActive     : boolean = true;
 glFullScreen : boolean = {not} true;

 glFPS    : DWORD; //  ... ;)


 function glInit: boolean;
 procedure glClose;
 function glDisplay(Width, Height, BPP, Freq: integer): boolean;
 procedure glSwap;
 procedure glSetProj(angle, zNear, zFar: single);
 procedure glSetOrtho(left, top, right, bottom: single);

 function glTex_Create(c, b: integer; Width, Height: integer; Data: pointer): TTexture;
 procedure glTex_Free(ID: TTexture);
 procedure glTex_Enable(ID: TTexture);
 procedure glTex_Disable;
 function glTex_Load(const FileName: string): TTexture;

 procedure glBlend(ID: Byte);
 procedure glLight_cfg(ID: integer);
 procedure glLight_Pos(ID: integer; Pos: TVector);
 procedure glLight_Color(ID: integer; R, G, B: single);

 function glFont_Create(const Name: string): TTexture;
 procedure glTextOut(font: TTexture; const Text: string; X, Y : Integer; sx : single = 1; sy: single = 1);
 procedure glBillBoard(Pos: TVector; Size: single);


implementation

uses
 TARGA;

type
 TManTex = record
  Tex  : TTexture;
  Name : string;
  Ref  : integer;
 end;

var
 cFPS   : DWORD = 0;  //     
 tFPS   : integer;    //     FPS
 CurTex : TTexture;   //  

 TexMan   : array of TManTex;
 FontBase : DWORD;

procedure Font_Init;
const
 cs = 1/16;
var
 i: DWORD;
 cx, cy: single;
begin
FontBase := glGenLists(256);
for i := 0 to 255 do
 begin
 glNewList(FontBase + i, GL_COMPILE);
 cx := (i mod 16)/16;
 cy := (i div 16)/16;

 glBegin(GL_QUADS);
  glTexCoord2f(cx,      -cy);      glVertex2f( 0,  0);
  glTexCoord2f(cx + cs, -cy);      glVertex2f(16,  0);
  glTexCoord2f(cx + cs, -cy - cs); glVertex2f(16, 16);
  glTexCoord2f(cx,      -cy - cs); glVertex2f( 0, 16);
 glEnd;

 glTranslatef(8, 0, 0);
 glEndList;
 end;
end;

procedure Font_Free; stdcall;
begin
glDeleteLists(FontBase, 256);
end;


(*********************)
(*      OpenGL       *)
(*********************)
//   //
function glInit: boolean;
var
 pfd     : PIXELFORMATDESCRIPTOR;
 iFormat : integer;
begin
Result := false;
DC := GetDC(Handle);

if DC = 0 then Exit;
FillChar(pfd, SizeOf(pfd), 0);
with pfd do
 begin
 nSize      := SizeOf(TPIXELFORMATDESCRIPTOR);
 nVersion   := 1;
 dwFlags    := PFD_DRAW_TO_WINDOW or
               PFD_SUPPORT_OPENGL or
               PFD_DOUBLEBUFFER;
 iPixelType := PFD_TYPE_RGBA;
 cColorBits := 32;
 cDepthBits := 16;
 iLayerType := PFD_MAIN_PLANE;
 end;

iFormat := ChoosePixelFormat(DC, @pfd);
if iFormat = 0 then
 Exit;

if not SetPixelFormat(DC, iFormat, @pfd) then
 Exit;

RC := wglCreateContext(DC);
if RC = 0 then Exit;
if not wglMakeCurrent(DC, RC) then
 Exit;

wglSwapIntervalEXT    := wglGetProcAddress('wglSwapIntervalEXT');
wglGetSwapIntervalEXT := wglGetProcAddress('wglGetSwapIntervalEXT');
glActiveTexture       := wglGetProcAddress('glActiveTexture');
glMultiTexCoord2fv    := wglGetProcAddress('glMultiTexCoord2fv');

glDepthFunc(GL_LEQUAL);
glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
glClearColor(0, 0, 0, 0);
glViewPort(0, 0, glWidth, glHeight);
Result := true;

glEnable(GL_COLOR_MATERIAL);
glEnable(GL_BLEND);
glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

Font_Init;
end;

procedure glClose;
begin
ChangeDisplaySettings(devmode(nil^), 0);
Font_Free;
wglMakeCurrent(DC, 0);
wglDeleteContext(RC);
ReleaseDC(Handle, DC);
DestroyWindow(Handle);
end;

function glDisplay(Width, Height, BPP, Freq: integer): boolean;
var
 dev  : TDeviceMode;
 res  : DWORD;
 bool : boolean;
begin
Result := false;
FillChar(dev, SizeOf(dev), 0);
dev.dmSize := sizeof(dev);
EnumDisplaySettings(nil, 0, dev);
with dev do
 begin
 dmPelsWidth        := Width;
 dmPelsHeight       := Height;
 dmBitsPerPel       := BPP;
 dmDisplayFrequency := Freq;
 dmFields := DM_BITSPERPEL or
             DM_PELSWIDTH  or
             DM_PELSHEIGHT or
             DM_DISPLAYFREQUENCY;
 res := ChangeDisplaySettings(dev, CDS_TEST or CDS_FULLSCREEN);
 if res = DISP_CHANGE_SUCCESSFUL then
  ChangeDisplaySettings(dev, CDS_FULLSCREEN);
 end;

if res <> DISP_CHANGE_SUCCESSFUL then
 begin
 bool := false;
 if Freq > 0 then
  bool := glDisplay(Width, Height, BPP, 0);
 if not bool then
  begin
  ChangeDisplaySettings(_devicemodeA(nil^), CDS_FULLSCREEN);
  Exit;
  end;
 end;

glWidth  := Width;
glHeight := Height;
glBPP    := BPP;
glFreq   := Freq;
MoveWindow(Handle, 0, 0, glWidth, glHeight, false);
Result := true;
end;

procedure glSwap;
var
 t : integer;
begin
if wglGetSwapIntervalEXT <> 0 then
 wglSwapIntervalEXT(1);
glFlush;
glFinish;
SwapBuffers(DC);
cFPS := cFPS + 1;
t := GetTimer;
if t - tFPS >= 1000 then
 begin
 tFPS  := t;
 glFPS := cFPS;
 cFPS  := 0;
 end;
end;

procedure glSetProj(angle, zNear, zFar: single);
begin
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluPerspective(angle, glWidth/glHeight, zNear, zFar);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
end;

procedure glSetOrtho(left, top, right, bottom: single);
begin
glMatrixMode(GL_PROJECTION);
glLoadIdentity;
gluOrtho2D(left, right, bottom, top);
glMatrixMode(GL_MODELVIEW);
glLoadIdentity;
end;

//   
function glTex_Create(c, b: integer; Width, Height: integer; Data: pointer): TTexture;
begin
glGenTextures(1, @Result);
glBindTexture(GL_TEXTURE_2D, Result);
glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
gluBuild2DMipmaps(GL_TEXTURE_2D, c, Width, Height, b, GL_UNSIGNED_BYTE, Data);
end;

function glTex_Load(const FileName: string): TTexture;
var
 TGA  : TGA_Header;
 i, j : integer;
begin
for i := 0 to Length(TexMan) - 1 do
 if FileName = TexMan[i].Name then
  begin
  Result := TexMan[i].Tex;
  inc(TexMan[i].Ref);
  Exit;
  end;

TGA := LoadTGA(FileName);
if TGA.Data = nil then
 Result := 0
else
 begin
 if TGA.BPP = 32 then
  begin
  i := $80E1;
  j := GL_RGBA8;
  end
 else
  begin
  i := $80E0;
  j := GL_RGB8;
  end;

 with TGA do
  begin
  Result := glTex_Create(j, i, Width, Height, Data);
  FreeMem(Data);
  end;
 i := Length(TexMan);
 SetLength(TexMan, i + 1);
 TexMan[i].Tex  := Result;
 TexMan[i].Name := FileName;
 TexMan[i].Ref  := 1;
 end;
end;

procedure glTex_Free(ID: TTexture);
var
 i, j : integer;
begin
if ID > 0 then
 begin
 j := -1;
 for i := 0 to Length(TexMan) - 1 do
  if TexMan[i].Tex = ID then
   begin
   j := i;
   dec(TexMan[i].Ref);
   break;
   end;

 if j < 0 then
  glDeleteTextures(1, @ID)
 else
  if TexMan[j].Ref <= 0 then
   begin
   for i := j to Length(TexMan) - 2 do
    TexMan[i] := TexMan[i + 1];
   SetLength(TexMan, Length(TexMan) - 1);
   glDeleteTextures(1, @ID)
   end;
 end;
end;

procedure glTex_Enable(ID: TTexture);
begin
if CurTex <> ID then
 begin
 if CurTex = 0 then
  glEnable(GL_TEXTURE_2D);
 glBindTexture(GL_TEXTURE_2D, ID);
 CurTex := ID;
 end;
end;

procedure glTex_Disable;
begin
glDisable(GL_TEXTURE_2D);
glBindTexture(GL_TEXTURE_2D, 0);
CurTex := 0;
end;

//   //
procedure glBlend(ID: Byte);
begin
if ID = 0 then
 glDisable(GL_BLEND)
else
 begin
 glEnable(GL_BLEND);
  case ID of
   1 : glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   2 : glBlendFunc(GL_SRC_ALPHA, GL_ONE);
   3 : glBlendFunc(GL_ZERO, GL_SRC_COLOR);
  end;
 end;
end;

procedure glLight_cfg(ID: integer);
const
 light_position : array [0..3] of single = (1, 1, 1, 0);
 white_light    : array [0..3] of single = (1, 1, 1, 1);
begin
glLightfv(ID, GL_POSITION, @light_position);
glLightfv(ID, GL_DIFFUSE,  @white_light);
glLightfv(ID, GL_SPECULAR, @white_light);
end;

procedure glLight_Pos(ID: integer; Pos: TVector);
var
 p : array [0..3] of single;
begin
p[0] := Pos.X;
p[1] := Pos.Y;
p[2] := Pos.Z;
p[3] := 1;
glLightfv(ID, GL_POSITION, @p);
end;

procedure glLight_Color(ID: integer; R, G, B: single);
var
 c : array [0..3] of single;
begin
c[0] := R;
c[1] := G;
c[2] := B;
c[3] := 1;
glLightfv(ID, GL_DIFFUSE,  @c);
glLightfv(ID, GL_SPECULAR, @c);
end;

// TEXT //
function glFont_Create(const Name: string): TTexture;
begin
Result := glTex_Load('data\textures\font\' + Name + '.tga');
end;

procedure glTextOut(font: TTexture; const Text: string; X, Y : Integer; sx : single = 1; sy: single = 1);
var
 s : string;
 i : integer;
 l : integer;
begin
glEnable(GL_ALPHA_TEST);
glAlphaFunc(GL_GEQUAL, 1/255);

glListBase(FontBase);
glTex_Enable(font);
glPushMatrix;
glTranslatef(X, Y, 0);
glScalef(sx, sy, 1);
s := Text;
l := Length(s);
for i := 1 to l do
 glCallLists(1, GL_UNSIGNED_BYTE, @s[i]);
glPopMatrix;
glDisable(GL_ALPHA_TEST);
end;

procedure glBillBoard(Pos: TVector; Size: single);
var
 mm     : array [0..15] of single;
 v      : array [0..3] of TVector;
 v1, v2 : TVector;
begin
glGetFloatv(GL_MODELVIEW_MATRIX, @mm);
v1 := V_Mult(Vector(mm[0], mm[4], mm[8]), Size);
v2 := V_Mult(Vector(mm[1], mm[5], mm[9]), Size);

v[0] := V_Add(Pos, Vector(-v1.X - v2.X, -v1.Y - v2.Y, -v1.Z - v2.Z));
v[1] := V_Add(Pos, Vector( v1.X - v2.X,  v1.Y - v2.Y,  v1.Z - v2.Z));
v[2] := V_Add(Pos, Vector( v1.X + v2.X,  v1.Y + v2.Y,  v1.Z + v2.Z));
v[3] := V_Add(Pos, Vector(-v1.X + v2.X, -v1.Y + v2.Y, -v1.Z + v2.Z));

glBegin(GL_QUADS);
 glTexCoord2f(0, 1); glVertex3fv(@v[0]);
 glTexCoord2f(1, 1); glVertex3fv(@v[1]);
 glTexCoord2f(1, 0); glVertex3fv(@v[2]);
 glTexCoord2f(0, 0); glVertex3fv(@v[3]);
glEnd;
end;

end.

