unit l_map;

interface

uses
 Windows, OpenGL,
 l_math, l_gl, l_player;

type
 TMapHeader = record
  Magic     : array [0..3] of Char;
  MeshCount : WORD;
  ObjCount  : WORD;
  TexCount  : WORD;
  TexName   : array of string;
 end;

 TMeshHeader = record
  Alpha     : Byte;
  TexID     : WORD;
  Faces     : WORD;
  Vertices  : WORD;
  TexCoords : WORD;
 end;

 TMeshFace = packed array [0..2] of WORD;

 TMesh = record
  Header   : TMeshHeader;
  MaxDist  : single;
  Center   : TVector;
  Face     : packed array of TMeshFace;
  Vertex   : packed array of TVector;
  TexFace  : packed array of TMeshFace;
  TexCoord : packed array of TVector2D;
  Normal   : packed array of TVector;
  Plane    : packed array of TPlane;
 end;

 TObj = record
  ID  : Byte;
  Pos : TVector;
 end;

 TMap = object
  //protected
   Header : TMapHeader;
   Mesh   : array of TMesh;
   Obj    : array of TObj;
   TexID  : array of TTexture;
  //public
   Player : array of PPlayer;
   function pl_add: PPlayer;
   procedure Clear;
   procedure LoadFromFile(const FileName: string);
   procedure Render;
   procedure Update;
   function Collision(var Pos, Speed: TVector; R: single): boolean;
 end;

const
 MAP_GRAVITY = 9.81/1000;

implementation

uses
 l_game;

function TMap.pl_add: PPlayer;
var
 i : integer;
begin
SetLength(Player, Length(Player) + 1);
New(Player[Length(Player) - 1]);
Result := Player[Length(Player) - 1];
Result^.Init;
for i := 0 to Header.ObjCount - 1 do
 if Obj[i].ID = 0 then
  begin
  Result^.Pos := Obj[i].Pos;
  break;
  end;
end;

procedure TMap.Clear;
var
 i : integer;
begin
//  //
//   
for i := 0 to integer(Header.TexCount) - 1 do
 glTex_Free(TexID[i]);
TexID := nil;
// 
for i := 0 to integer(Header.MeshCount) - 1 do
 with Mesh[i] do
  begin
  Face     := nil;
  Vertex   := nil;
  TexFace  := nil;
  TexCoord := nil;
  Plane    := nil;
  end;
Mesh := nil;
// 
Obj := nil;
end;

procedure TMap.LoadFromFile(const FileName: string);
var
 b    : Byte;
 i, j : integer;
 F    : File of Byte;
 d    : single;
 m_t  : TMesh;
begin
Clear;

AssignFile(F, 'data/maps/' + FileName + '.xmp');
Reset(F);
with Header do
 begin
 BlockRead(F, Magic, SizeOf(Magic));
 if Magic <> 'XMAP' then
  begin
  CloseFile(F);
  Exit;
  end;
 BlockRead(F, MeshCount, SizeOf(MeshCount));
 BlockRead(F, ObjCount, SizeOf(ObjCount));
 BlockRead(F, TexCount, SizeOf(TexCount));
 SetLength(TexName, TexCount);

 for i := 0 to TexCount - 1 do
  begin
  BlockRead(F, b, 1);
  SetLength(TexName[i], b);
  BlockRead(F, TexName[i][1], b);
  end;

 SetLength(Mesh, MeshCount);
 for i := 0 to MeshCount - 1 do
  with Mesh[i] do
   begin
   BlockRead(F, Header, SizeOF(Header));
   SetLength(Face, Header.Faces);
   SetLength(Vertex, Header.Vertices);
   SetLength(TexFace, Header.Faces);
   SetLength(TexCoord, Header.TexCoords);
   BlockRead(F, Face[0], SizeOf(TMeshFace) * Header.Faces);
   BlockRead(F, Vertex[0], SizeOf(TVector) * Header.Vertices);
   BlockRead(F, TexFace[0], SizeOf(TMeshFace) * Header.Faces);
   BlockRead(F, TexCoord[0], SizeOf(TVector2D) * Header.TexCoords);
   end;
 SetLength(Obj, ObjCount);
 for i := 0 to ObjCount - 1 do
  BlockRead(F, Obj[i], SizeOf(TObj));
 end;
CloseFile(F);

//  
SetLength(TexID, Header.TexCount);
for i := 0 to Header.TexCount - 1 do
 TexID[i] := glTex_Load('data/textures/' + Header.TexName[i] + '.tga');

//   !
for i := 0 to Header.MeshCount - 1 do
 if Mesh[i].Header.Alpha < 255 then
  for j := Header.MeshCount - 1 downto i + 1 do
   if Mesh[j].Header.Alpha = 255 then
    begin
    m_t := Mesh[j];
    Mesh[j] := Mesh[i];
    Mesh[i] := m_t; 
    break;
    end;

//    
for i := 0 to Header.MeshCount - 1 do
 with Mesh[i] do
  begin
  //   
  SetLength(Plane, Header.Faces);
  for j := 0 to Header.Faces - 1 do
   Plane[j] := P_Plane(Vertex[Face[j, 0]],
                       Vertex[Face[j, 1]],
                       Vertex[Face[j, 2]]);
  //  "" 
  Center := Vector(0, 0, 0);
  for j := 0 to Header.Vertices - 1 do
   Center := V_Add(Center, Vertex[j]);
  Center := V_Div(Center, Header.Vertices);
  //       
  MaxDist := -1;
  for j := 0 to Header.Vertices - 1 do
   begin
   d := V_Distq(Center, Vertex[j]);
   if d > MaxDist then
    MaxDist := d;
   end;
  MaxDist := sqrt(MaxDist);
  end;
end;

procedure TMap.Render;
var
 i, j : integer;
begin
for i := 0 to Header.MeshCount - 1 do
 with Mesh[i] do
  begin
  glTex_Enable(TexID[Mesh[i].Header.TexID]);
  glColor4ub(255, 255, 255, Header.Alpha);
  glBegin(GL_TRIANGLES);
   for j := 0 to Header.Faces - 1 do
    if P_Classify(Camera.Pos, Plane[j]) > 0 then
     begin
     glNormal3fv(@Plane[j].Normal);
     glTexCoord2fv(@TexCoord[TexFace[j, 0]]); glVertex3fv(@Vertex[Face[j, 0]]);
     glTexCoord2fv(@TexCoord[TexFace[j, 1]]); glVertex3fv(@Vertex[Face[j, 1]]);
     glTexCoord2fv(@TexCoord[TexFace[j, 2]]); glVertex3fv(@Vertex[Face[j, 2]]);
     end;
  glEnd;
  end;
end;

procedure TMap.Update;
var
 i : integer;
 p : TVector;
begin
for i := 0 to Length(Player) - 1 do
 with Player[i]^ do
  begin
  Acc.Y := -MAP_GRAVITY;
  Update;
  //   
  p := Pos;
  onGround := Collision(p, Speed, PLAYER_RAD);
  //   
  p := V_Add(p, Vector(0, 1, 0));
  Collision(p, Speed, PLAYER_RAD);
  p := V_Sub(p, Vector(0, 1, 0));
  Pos := p;
  end;
end;

function TMap.Collision(var Pos, Speed: TVector; R: single): boolean;
var
 i, j   : integer;
 Dist   : single;
 offset : TVector;
 ins    : TVector;
begin
Result := false;
for i := 0 to Header.MeshCount - 1 do
 with Mesh[i] do
  if V_Dist(Center, Pos) <= MaxDist + R then
   for j := 0 to Header.Faces - 1 do
    if P_Classify(Pos, Plane[j]) > 0 then
     begin
     Dist := P_Dist(Pos, Plane[j]);
     if Dist < R then
      begin
      offset := V_Mult(Plane[j].Normal, R - Dist);
      ins    := V_Sub(Pos, V_Mult(Plane[j].Normal, Dist));
      if P_InsideTri(ins,
                     Vertex[Face[j, 0]],
                     Vertex[Face[j, 1]],
                     Vertex[Face[j, 2]]) {or
  		    EdgeSphereCollision(Pos, Vertex[Face[j, 0]],
                     Vertex[Face[j, 1]],
                     Vertex[Face[j, 2]], R/2)} then
       begin
       Pos := V_Add(Pos, offset);
       if abs(Pos.Y - ins.Y) > R/2 then
        begin
        Speed.Y := 0;
        if Pos.Y > ins.Y then
         Result := true;
        end;
       end;
      end;
     end;
end;

end.
