unit l_math;
(*******************************)
(* TRacer math libary          *)
(*******************************)
(* Created by XProger          *)
(* begin : 16.04.2005          *)
(* end   : 19.04.2005          *)
(*******************************)
(* mail : XProger@list.ru      *)
(*******************************)
interface

uses
 Windows;

const
 rad2deg = 180/pi;
 deg2rad = pi/180;

type
 TVector2D = record
  X, Y : single;
 end;

 PVector = ^TVector;
 TVector = record
  X, Y, Z : single;
 end;

 TPlane = record
  Normal : TVector;
  V      : TVector;
  D      : single;
 end;

 function max(x, y: single): single;
 function min(x, y: single): single;

 function ArcTan2(const Y, X: single): single;
 function ArcCos(const X: single): single;
// 3D
 function Vector(X, Y, Z: single): TVector;
 function V_Add(v1, v2: TVector): TVector;
 function V_Sub(v1, v2: TVector): TVector;
 function V_Mult(v: TVector; d: single): TVector;
 function V_Div(v: TVector; d: single): TVector;
 function V_Dist(v1, v2: TVector): single;
 function V_Distq(v1, v2: TVector): single;
 function V_Length(v: TVector): single;
 function V_Normalize(v: TVector): TVector;
 function V_Dot(v1, v2: TVector): single;
 function V_Cross(v1, v2: TVector): TVector;
 function V_Angle(v1, v2: TVector): single;
 function V_Interpolate(v1, v2: TVector; k: single): TVector;
// 2D
 function Vector2D(X, Y: single): TVector2D;
 function V_Rotate2D(v: TVector2D; ang: single): TVector2D;
// Plane
 function P_Normal(v1, v2, v3: TVector): TVector;
 function P_Plane(v1, v2, v3: TVector): TPlane;
 function P_Dist(Pos: TVector; Plane: TPlane): single;
 function P_Classify(Pos: TVector; Plane: TPlane): single;
 function P_InsideTri(v, v1, v2, v3: TVector): boolean;
///////
function EdgeSphereCollision(Center: TVector; v1, v2, v3: TVector; Radius: Single): Boolean;
function ClosestPointOnLine(vA, vB, Point: TVector): TVector;

 
implementation


function max(x, y: single): single;
begin
if x > y then
 Result := x
else
 Result := y;
end;

function min(x, y: single): single;
begin
if x < y then
 Result := x
else
 Result := y;
end;

(***      ***)

// 
function ArcTan2(const Y, X: single): single;
asm
 FLD     Y
 FLD     X
 FPATAN
 FWAIT
end;

function ArcCos(const X: single): single;
begin
Result := ArcTan2(Sqrt(1 - X * X), X);
end;

(***      ***)

//  
function Vector(X, Y, Z: single): TVector;
begin
Result.X := X;
Result.Y := Y;
Result.Z := Z;
end;

//  
function V_Add(v1, v2: TVector): TVector;
begin
Result.X := v1.X + v2.X;
Result.Y := v1.Y + v2.Y;
Result.Z := v1.Z + v2.Z;
end;

//  
function V_Sub(v1, v2: TVector): TVector;
begin
Result.X := v1.X - v2.X;
Result.Y := v1.Y - v2.Y;
Result.Z := v1.Z - v2.Z;
end;

function V_Mult(v: TVector; d: single): TVector;
begin
Result.X := v.X * d;
Result.Y := v.Y * d;
Result.Z := v.Z * d;
end;

function V_Div(v: TVector; d: single): TVector;
begin
if d <> 0 then
 Result := Vector(v.X/d, v.Y/d, v.Z/d)
else
 Result := v;
end;

//    
function V_Dist(v1, v2: TVector): single;
begin
Result := sqrt(sqr(v2.X - v1.X) +
               sqr(v2.Y - v1.Y) +
               sqr(v2.Z - v1.Z));
end;

//     
function V_Distq(v1, v2: TVector): single;
begin
Result := sqr(v2.X - v1.X) +
          sqr(v2.Y - v1.Y) +
          sqr(v2.Z - v1.Z);
end;

function V_Length(v: TVector): single;
begin
Result := sqrt(sqr(v.X) + sqr(v.Y) + sqr(v.Z));
end;

function V_Normalize(v: TVector): TVector;
begin
Result := V_Mult(v, 1/V_Length(v));
end;

function V_Dot(v1, v2: TVector): single;
begin
Result := v1.X * v2.X + v1.Y * v2.Y + v1.Z * v2.Z;
end;

function V_Cross(v1, v2: TVector): TVector;
begin
Result.X := v1.Y * v2.Z - v1.Z * v2.Y;
Result.Y := v1.Z * v2.X - v1.X * v2.Z;
Result.Z := v1.X * v2.Y - v1.Y * v2.X;
end;

function V_Angle(v1, v2: TVector): single;
begin
Result := ArcCos(V_Dot(V_Normalize(v1), V_Normalize(v2)));
end;

function V_Interpolate(v1, v2: TVector; k: single): TVector;
begin
Result.X := v1.X + (v2.X - v1.X)/k;
Result.Y := v1.Y + (v2.Y - v1.Y)/k;
Result.Z := v1.Z + (v2.Z - v1.Z)/k;
end;

// 2D
function Vector2D(X, Y: single): TVector2D;
begin
Result.X := X;
Result.Y := Y;
end;

function V_Rotate2D(v: TVector2D; ang: single): TVector2D;
var
 l : single;
begin
l := sqrt(sqr(v.X) + sqr(v.Y));
Result.X := cos(ang) * l;
Result.Y := sin(ang) * l;
end;

// Plane
function P_Normal(v1, v2, v3: TVector): TVector;
begin
Result := V_Normalize(V_Cross(V_Sub(v2, v1), V_Sub(v3, v2)));
end;

function P_Plane(v1, v2, v3: TVector): TPlane;
begin
Result.Normal := P_Normal(v1, v2, v3);
Result.D      := -V_Dot(Result.Normal, v1);
Result.V      := v1;
end;

function P_Dist(Pos: TVector; Plane: TPlane): single;
begin
Result := (V_Dot(Plane.Normal, Pos) + Plane.D);
end;

function P_Classify(Pos: TVector; Plane: TPlane): single;
begin
Result := V_Dot(Plane.Normal, V_Sub(Pos, Plane.V));
end;

function P_InsideTri(v, v1, v2, v3: TVector): boolean;
var
 i      : Integer;
 Angle  : Single;
 vec    : array [0..2] of TVector;
begin
Result := False;
vec[0] := V_Sub(v1, v);
vec[1] := V_Sub(v2, v);
vec[2] := V_Sub(v3, v);
Angle := 0;
for i := 0 to 2 do
 Angle := Angle + V_Angle(vec[i], vec[(i + 1) mod 3]);
if Angle >= (0.9 * 2 * pi) then
 Result := true;
end;

///////
function ClosestPointOnLine(vA, vB, Point: TVector): TVector;
var
 Vector1, Vector2, Vector3: TVector;
 d, t: Single;
begin
Vector1 := V_Sub(Point, vA);
Vector2 := V_Normalize(V_Sub(vB, vA));
d := V_Dist(vA, vB);
t := V_Dot(Vector2, Vector1);

if t <= 0 then
 begin
 Result := vA;
 Exit;
 end;

if t >= d then
 begin
 Result := vB;
 Exit;
 end;

Vector3 := V_Mult(Vector2, t);
Result := V_Add(vA, Vector3);
end;

function EdgeSphereCollision(Center: TVector; v1, v2, v3: TVector; Radius: Single): Boolean;
var i: Integer;
    Point: TVector;
    Distance: Single;
 v : array [0..2] of TVector;
begin
v[0] := v1;
v[1] := v2;
v[2] := v3;
for i := 0 to 2 do
 begin
 Point := ClosestPointOnLine(v[i], v[(i + 1) mod 3], Center);
 Distance := V_Dist(Point, Center);
 if Distance < Radius then
  begin
  Result := true;
  Exit;
  end;
 end;
Result := false;
end;

end.
